/**
 * Data collector for the epidemiology application
 * This collects data from a mote and save it to a local storage. 
 *
 * @author Jung Woo Lee <jungwoo@gmail.com>
 * @date September 18 2009
 */

import java.io.*;
import java.util.*;
import java.text.*;

import net.tinyos.message.*;
import net.tinyos.packet.*;
import net.tinyos.util.*;

public class BaseStation implements MessageListener {

  /* member variable */
  private int m_state;         //current state in a state machine.
  private int m_currentNodeID; //current node ID from which data is retrieved
  private int m_retrans_num;   //number of remaining retransmission
  private MoteIF moteIF;
  private Timer m_waitTimer;
  private boolean m_isTimerOn;

  public class LogEntry {
      private int m_source;
      private int  m_seqno;
      private short m_rssi;
      private int m_myseqno;
      public void setSource(int source){ m_source = source;}
      public void setSeqno(int seqno){ m_seqno = seqno;}
      public void setRssi(short rssi){ m_rssi = rssi;}
      public void setMyseq(int my_seqno) { m_myseqno = my_seqno;} 
  }
  /* variable and constant for data entries */
  public static final int MAX_ENTRY = 210000; // (210000 ) * 5 bytes =  1.001 MBytes
  private LogEntry[] m_log_entry = new LogEntry[MAX_ENTRY];
  private int m_num_entry;

  /* state constants */
  public static final int STATE_START			= 1;
  public static final int STATE_PINGING 		= 2;
  public static final int STATE_WAIT_FOR_COLLECTION 	= 3;
  public static final int STATE_START_COLLECTION    	= 4;
  public static final int STATE_COLLECTING    		= 5;
  public static final int STATE_END			= 6;

  /* constants for timer */
  public static final int PINGING_INTERVAL  = 250;          //time interval between pinging
  public static final int RETRIEVE_INTERVAL = 250;          //time interval between pinging
  //WAIT_BEFORE_COLLECTION = 512,    //wait until other nodes are calm down

  /* AM IDs */
  public static final int AM_CONTROL_PACKET_T = 172;
  public static final int AM_DATA_T           = 173;

  /* types of network packets */
  public static final short BEACON = 1;
  public static final short PING_REQUEST = 2;
  public static final short PING_REPLY = 3;
  public static final short RETRIEVE = 4;
  public static final short DATA = 5;
  public static final short END = 6;
  public static final short TEST = 7;

  /* range of node ids */
  public static final int START_NODE_ID = 0;
  public static final int END_NODE_ID = 10000;

  /* other constants */
  public static final int MAX_CONTROL_RETRANSMISSIONS = 5; //maximum number of retransmissions of the "start_transfer" packet
  public static final int MAX_DATA_INTERVAL = 2000;        //maximum allowed interval between data transfers. If no packet is received in this interval, 


  public BaseStation(MoteIF moteIF) {
    this.moteIF = moteIF;
    this.moteIF.registerListener(new control_packet_t(), this);
    this.moteIF.registerListener(new data_t(), this);
    init();
  }

  private void enter_pinging_state(){
     m_state = STATE_PINGING;
     System.out.println("entering pinging state");
     send_ping();
  }

  private void send_ping(){
    control_packet_t  ctrl_pkt = new control_packet_t();
    ctrl_pkt.set_type(PING_REQUEST);
    ctrl_pkt.set_startid(START_NODE_ID);
    ctrl_pkt.set_endid(END_NODE_ID);
    
    if (m_isTimerOn){
        m_waitTimer.cancel();
        m_isTimerOn = false;
        m_waitTimer = new Timer();
    }
    m_waitTimer.schedule( new TimerFired(), PINGING_INTERVAL);
    m_isTimerOn = true;
    System.out.println("sending a ping");
    try {
       	moteIF.send(moteIF.TOS_BCAST_ADDR,ctrl_pkt);
    }
    catch (IOException exception) {
    	System.err.println("Exception thrown in send_ping()");
        System.err.println(exception);
        if(m_isTimerOn){
            m_waitTimer.cancel();
            m_isTimerOn = false;
            m_waitTimer = new Timer();
        }
        send_ping();
    }
  }

  private void initiate_collection(){
    //create the "transfer_start" message
    control_packet_t ctrl_pkt = new control_packet_t();
    ctrl_pkt.set_type(RETRIEVE);
    if (m_isTimerOn){
        m_waitTimer.cancel();
        m_isTimerOn = false;
        m_waitTimer = new Timer();
    }
    m_waitTimer.schedule( new TimerFired(), RETRIEVE_INTERVAL);
    m_isTimerOn = true;
    try {
       	moteIF.send(m_currentNodeID, ctrl_pkt);
        System.out.println("sending a retrieve message to node "+m_currentNodeID);
    }
    catch (IOException exception) {
    	System.err.println("Exception thrown in initiate_collection()");
        System.err.println(exception);
        initiate_collection();
    }
  }

  class TimerFired extends TimerTask {
      public void run(){
          m_isTimerOn = false;
          switch (m_state){
              case STATE_PINGING:
                  send_ping();
                  break;
              case STATE_START_COLLECTION:
                  if (--m_retrans_num !=0)
                      initiate_collection();
                  else
                      enter_pinging_state();//no neighbor found
                  break;
              case STATE_COLLECTING:
                  enter_pinging_state();
                  break;
          }
      }
  }

  private void init()
  {
    m_waitTimer = new Timer();
    m_isTimerOn = false;
    for (int i=0; i< MAX_ENTRY; i++){
       m_log_entry[i] = new LogEntry();
    }
    m_num_entry = 0;
    System.out.println("Basestation initialized");
    enter_pinging_state();
   
  }

  private void savedata()
  {
  	try{
	  SimpleDateFormat df = new SimpleDateFormat("yyyy.MM.dd_HH.mm.ss");
	  String datestr = df.format(new Date());
          File outFile = new File ("data/node-"+m_currentNodeID+"-" + datestr);
  	  FileWriter fWriter = new FileWriter(outFile);
	  PrintWriter pWriter = new PrintWriter(fWriter);
	  for (int i=0; i< m_num_entry; i++){
	     //pWriter.printf("%8d %8d %4d", m_log_entry[i].m_source, m_log_entry[i].m_seqno, m_log_entry[i].m_rssi);
	     pWriter.printf("%d %d %d %d", m_log_entry[i].m_source, m_log_entry[i].m_seqno, m_log_entry[i].m_rssi, m_log_entry[i].m_myseqno);
             pWriter.printf("\n");
	  }
          System.out.println("Data is saved in node-"+m_currentNodeID + "-" +datestr);
          fWriter.close();
        }
        catch (Exception e){
	  System.out.println(e.getMessage());
 	  e.printStackTrace();
        }
        System.exit(0);
  }
  
  private void recordLogEntry(int source, int seqno, short rssi, int my_seqno) {
    System.out.println("source: "+source+" seqno: "+seqno+" rssi: "+rssi+" myseq: "+my_seqno);
    m_log_entry[m_num_entry].setSource(source);
    m_log_entry[m_num_entry].setSeqno(seqno);
    m_log_entry[m_num_entry].setRssi(rssi);
    m_log_entry[m_num_entry].setMyseq(my_seqno);
    m_num_entry++;
  }

  public void messageReceived(int to, Message message) {
    switch(m_state){
       //In PINGING state, if we receive a ping_reply,  
       //we wait for a while to suppress the subsequent inconing replies that could collide with data transmissions
       case STATE_PINGING:
           if (message.amType() == AM_CONTROL_PACKET_T){
               control_packet_t ctl_pkt = (control_packet_t)message;
               if (ctl_pkt.get_type() == PING_REPLY){
                   m_currentNodeID  = ctl_pkt.get_sourceid(); 	//get the sender id of ping_reply
                   System.out.println("A ping-reply was received from node "+m_currentNodeID);
                   m_state = STATE_START_COLLECTION;  	
                   m_retrans_num = MAX_CONTROL_RETRANSMISSIONS;
                   initiate_collection();
               }
           }
           break;
       //In START_COLLECTION state, we send RETRIVE control packet until we receive the first data packet
       case STATE_START_COLLECTION:
           if (message.amType() == AM_DATA_T){
               data_t data_pkt = (data_t)message;
               if (data_pkt.get_type() == DATA){
                   m_state = STATE_COLLECTING;
                   System.out.println("the first data packet was received from node "+m_currentNodeID);
                   Dump.printPacket(System.out, message.dataGet());
                   System.out.println();
                   for (int i=0; i< data_pkt.get_num_entries(); i++){
		       int source = (data_pkt.get_entries_source())[i];
                       int seqno  = (data_pkt.get_entries_seqno())[i];
                       short rssi  =  (data_pkt.get_entries_rssi()[i]);
                       int my_seqno = (data_pkt.get_entries_my_seqno()[i]);
                       recordLogEntry(source, seqno, rssi, my_seqno);
                   }
                   if (m_isTimerOn){
                       m_waitTimer.cancel();
                       m_isTimerOn = false;
                       m_waitTimer = new Timer();
                   }
                   m_waitTimer.schedule( new TimerFired(), MAX_DATA_INTERVAL);
                   m_isTimerOn = true;
               }
           }
           break;
       //In COLLECTING state, we start a timer upon reception of every data packet. If the timer expires, we assume that connection to the node is down and finish the transfer
       //If we receive the END control packet, it finishes the connection
       case STATE_COLLECTING:
           if (message.amType() == AM_DATA_T){
               data_t data_pkt = (data_t)message;
               if (data_pkt.get_type() == DATA){
                   Dump.printPacket(System.out, message.dataGet());
                   System.out.println();
                   for (int i=0; i< data_pkt.get_num_entries(); i++){
		       int source = (data_pkt.get_entries_source())[i];
                       int seqno  = (data_pkt.get_entries_seqno())[i];
                       short rssi  =  (data_pkt.get_entries_rssi()[i]);
                       int my_seqno = (data_pkt.get_entries_my_seqno()[i]);
                       recordLogEntry(source, seqno, rssi, my_seqno);
                   }
                   if (m_isTimerOn){
                       m_waitTimer.cancel();
 		       m_isTimerOn = false;
                       m_waitTimer = new Timer();
                   }
                   m_waitTimer.schedule( new TimerFired(), MAX_DATA_INTERVAL);
 	  	   m_isTimerOn = true;
               }else{
                   System.out.println("A unexpected packet was received!");
               } 
           }
           else if (message.amType() == AM_CONTROL_PACKET_T){
               control_packet_t ctl_pkt = (control_packet_t)message;
               if (ctl_pkt.get_type() == END){
                   System.out.println("The end packet was received from node "+m_currentNodeID);
                   if (m_isTimerOn){
                       m_waitTimer.cancel();
                       m_isTimerOn = false;
                       m_waitTimer = new Timer();
                   }
                   //enter_pinging_state();
                   savedata();
               }
           }
           break;
    }
  }
  
  private static void usage() {
    System.err.println("usage: BaseStation [-comm <source>]");
  }
  
/*
  public static short flip(short y)
  {
	return y;
  }
*/

  public static short flip(short y)
  {
    int x = (int)y;
    int high = (x << 8) & 0x0000FF00;
    int low = ( x >> 8) & 0x000000FF;
    int z = high | low;
    short w = (short)z;
    return w;
  }

  public static void main(String[] args) throws Exception {
    String source = null;
    int i = 0;
    short dest = 0;

    PhoenixSource phoenix;
    
    source = args[1];
    if (source == null) {
      phoenix = BuildSource.makePhoenix(PrintStreamMessenger.err);
      System.out.println("source is null.");
    }
    else {
      phoenix = BuildSource.makePhoenix(source, PrintStreamMessenger.err);
    }

    MoteIF mif = new MoteIF(phoenix);
    BaseStation serial = new BaseStation(mif);
  }

}
